@@ -13,7 +13,7 @@ print.clPlatformID <- function(x, ...) {
 oclPlatforms <- function() .Call("ocl_platforms")
 oclDevices <- function(platform = oclPlatforms()[[1]], type="default") .Call("ocl_devices", platform, type)
 oclSimpleKernel <- function(device, name, code, precision=c("single","double")) .Call("ocl_ez_kernel", device, name, code, match.arg(precision))
-oclRun <- function(kernel, size, ...) .External("ocl_call", kernel, size, ...)
+oclRun <- function(kernel, size, ..., native.result=FALSE) .External("ocl_call", kernel, size, native.result, ...)
 
 oclInfo <- function(item) UseMethod("oclInfo")
 oclInfo.clDeviceID <- function(item) .Call("ocl_get_device_info", item)
