@@ -6,6 +6,7 @@
 #include <CL/opencl.h>
 #endif
 
+#define USE_RINTERNALS 1
 #include <Rinternals.h>
 
 void ocl_err(const char *str) {
@@ -263,23 +264,27 @@ SEXP ocl_ez_kernel(SEXP device, SEXP k_name, SEXP code, SEXP prec) {
     }
 
     kernel = clCreateKernel(program, CHAR(STRING_ELT(k_name, 0)), &err);
-    if (!kernel) {
-	clReleaseProgram(program);
+    clReleaseProgram(program);
+    if (!kernel)
 	ocl_err("clCreateKernel");
-    }
 
-    /* FIXME: do we need to retain/release queue and program? */
     {
 	SEXP sk = PROTECT(mkKernel(kernel));
 	Rf_setAttrib(sk, Rf_install("device"), device);
 	Rf_setAttrib(sk, Rf_install("precision"), prec);
+	Rf_setAttrib(sk, Rf_install("context"), sctx);
 	UNPROTECT(2); /* sk + context */
 	return sk;
     }
 }
 
+
+/*--- generic blockwise argument list that can be freed in one go ---*/
+typedef void (*afin_t)(void*);
+
 struct arg_chain {
     struct arg_chain *next;
+    afin_t fin;
     int args, size;
     void *arg[1];
 };
@@ -291,6 +296,7 @@ static struct arg_chain *arg_alloc(struct arg_chain *parent, int size) {
     c->next = 0;
     c->size = size;
     c->args = 0;
+    c->fin = 0;
     if (parent)
 	parent->next = c;
     return c;
@@ -307,30 +313,61 @@ static struct arg_chain *arg_add(struct arg_chain *where, void *arg) {
     return where;
 }
 
-static void arg_free(struct arg_chain *chain) {
+static void arg_free(struct arg_chain *chain, afin_t fin) {
     int i, n = chain->args;
     if (chain->next)
-	arg_free(chain->next);
+	arg_free(chain->next, fin);
     for (i = 0; i < n; i++)
-	free(chain->arg[i]);
+	if (fin) fin(chain->arg[i]); else free(chain->arg[i]);
     free(chain);
 }
 
+#if 0 /* unused - we use it as part of the call context instead */
 static void free_protected_args(SEXP o) {
-    arg_free((struct arg_chain*)R_ExternalPtrAddr(o));
+    arg_free((struct arg_chain*)R_ExternalPtrAddr(o), 0);
 }
 
-static SEXP protected_args(struct arg_chain *chain) {
+static SEXP protected_args(struct arg_chain *chain, afin_t fin) {
     SEXP res = R_MakeExternalPtr(chain, R_NilValue, R_NilValue);
+    chain->fin = fin;
     R_RegisterCFinalizerEx(res, free_protected_args, TRUE);
     return res;
 }
+#endif
+
+/* in order to clean up all the temporary OpenCL objects we
+   keep them in this structure which we allocate as an external
+   pointer so it will be clean up either by the garbage collector
+   in case of an error or by us at the end of the call.
+   This is rather important for memory obejcts since GPUs have
+   only limited memory.
+*/
+typedef struct ocl_call_context {
+    cl_mem output;
+    cl_command_queue commands;
+    void *float_out;
+    struct arg_chain *float_args, *mem_objects;
+} ocl_call_context_t;
+
+static void ocl_call_context_fin(SEXP context) {
+    ocl_call_context_t *ctx = (ocl_call_context_t*) R_ExternalPtrAddr(context);
+    if (ctx) {
+	if (ctx->output) clReleaseMemObject(ctx->output);
+	if (ctx->float_args) arg_free(ctx->float_args, 0);
+	if (ctx->float_out) free(ctx->float_out);
+	if (ctx->mem_objects) arg_free(ctx->mem_objects, (afin_t) clReleaseMemObject);
+	if (ctx->commands) clReleaseCommandQueue(ctx->commands);
+	free(ctx);
+	CAR(context) = 0; /* this allows us to call the finalizer manually */
+    }
+}
 
 SEXP ocl_call(SEXP args) {
     struct arg_chain *float_args = 0;
-    int on, an = 0, ftype = FT_DOUBLE, ftsize;
+    ocl_call_context_t *occ;
+    int on, an = 0, ftype = FT_DOUBLE, ftsize, ftres;
     size_t global;
-    SEXP ker = CADR(args), olen, arg, res;
+    SEXP ker = CADR(args), olen, arg, res, octx;
     cl_kernel kernel = getKernel(ker);
     cl_context context;
     cl_command_queue commands;
@@ -345,25 +382,33 @@ SEXP ocl_call(SEXP args) {
     if (TYPEOF(res) == STRSXP && LENGTH(res) == 1 && CHAR(STRING_ELT(res, 0))[0] != 'd')
 	ftype = FT_SINGLE;
     ftsize = (ftype == FT_DOUBLE) ? sizeof(double) : sizeof(float);
-    if (ftype == FT_SINGLE) /* need conversions, protect floats buffer */
-	PROTECT(protected_args(float_args = arg_alloc(0, 32)));
     olen = CAR(args);
     args = CDR(args);
     on = Rf_asInteger(olen);
     if (on < 0)
 	Rf_error("invalid output length");
-    output = clCreateBuffer(context, CL_MEM_WRITE_ONLY, ftsize * on, NULL, &err);
+    ftres = (Rf_asInteger(CAR(args)) == 1) ? 1 : 0;
+    if (ftype != FT_SINGLE) ftres = 0;
+    args = CDR(args);    
+    occ = (ocl_call_context_t*) calloc(1, sizeof(ocl_call_context_t));
+    if (!occ) Rf_error("unable to allocate ocl_call context");
+    octx = PROTECT(R_MakeExternalPtr(occ, R_NilValue, R_NilValue));
+    R_RegisterCFinalizerEx(octx, ocl_call_context_fin, TRUE);
+
+    occ->output = output = clCreateBuffer(context, CL_MEM_WRITE_ONLY, ftsize * on, NULL, &err);
     if (!output)
 	Rf_error("failed to create output buffer of %d elements via clCreateBuffer (%d)", on, err);
     if (clSetKernelArg(kernel, an++, sizeof(cl_mem), &output) != CL_SUCCESS)
 	Rf_error("failed to set first kernel argument as output in clSetKernelArg");
     if (clSetKernelArg(kernel, an++, sizeof(on), &on) != CL_SUCCESS)
 	Rf_error("failed to set second kernel argument as output length in clSetKernelArg");
-    commands = clCreateCommandQueue(context, device_id, 0, &err);
+    occ->commands = commands = clCreateCommandQueue(context, device_id, 0, &err);
     if (!commands)
 	ocl_err("clCreateCommandQueue");
+    if (ftype == FT_SINGLE) /* need conversions, create floats buffer */
+	occ->float_args = float_args = arg_alloc(0, 32);
     while ((arg = CAR(args)) != R_NilValue) {
-	int n;
+	int n, ndiv = 1;
 	void *ptr;
 	size_t al;
 	
@@ -394,6 +439,12 @@ SEXP ocl_call(SEXP args) {
 	    ptr = LOGICAL(arg);
 	    al = sizeof(int);
 	    break;
+	case RAWSXP:
+	    if (inherits(arg, "clFloat")) {
+		ptr = RAW(arg);
+		ndiv = al = sizeof(float);
+		break;
+	    }
 	default:
 	    Rf_error("only numeric or logical kernel arguments are supported");
 	    /* no-ops but needed to make the compiler happy */
@@ -401,15 +452,21 @@ SEXP ocl_call(SEXP args) {
 	    al = 0;
 	}
 	n = LENGTH(arg);
+	if (ndiv != 1) n /= ndiv;
 	if (n == 1) {/* scalar */
 	    if (clSetKernelArg(kernel, an++, al, ptr) != CL_SUCCESS)
 		Rf_error("Failed to set scalar kernel argument %d (size=%d)", an, al);
 	} else {
-	    cl_mem input = clCreateBuffer(context,  CL_MEM_READ_ONLY,  al * n, NULL, NULL);
+	    cl_mem input = clCreateBuffer(context,  CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR,  al * n, ptr, &err);
 	    if (!input)
-		Rf_error("Unable to create buffer for vector argument %d", an);
+		Rf_error("Unable to create buffer (%d elements, %d bytes each) for vector argument %d (oclError %d)", n, al, an, err);
+	    if (!occ->mem_objects)
+		occ->mem_objects = arg_alloc(0, 32);
+	    arg_add(occ->mem_objects, input);
+#if 0 /* we used this before CL_MEM_USE_HOST_PTR */
 	    if (clEnqueueWriteBuffer(commands, input, CL_TRUE, 0, al * n, ptr, 0, NULL, NULL) != CL_SUCCESS)
 		Rf_error("Failed to transfer data (%d elements) for vector argument %d", n, an);
+#endif
 	    if (clSetKernelArg(kernel, an++, sizeof(cl_mem), &input) != CL_SUCCESS)
 		Rf_error("Failed to set vector kernel argument %d (size=%d, length=%d)", an, al, n);
 	    /* clReleaseMemObject(input); */
@@ -422,24 +479,41 @@ SEXP ocl_call(SEXP args) {
 	Rf_error("Error during kernel execution");
     clFinish(commands);
 
-    res = Rf_allocVector(REALSXP, on);
-    if (ftype == FT_SINGLE) { /* float - need a temporary buffer */
-	float *fr = (float*) malloc(sizeof(float) * on);
-	double *r = REAL(res);
-	int i;
-	if (!fr)
-	    Rf_error("unable to allocate memory for temporary single-precision output buffer");
-	arg_add(float_args, fr);
-	if (clEnqueueReadBuffer( commands, output, CL_TRUE, 0, sizeof(float) * on, fr, 0, NULL, NULL ) != CL_SUCCESS)
-	    Rf_error("Unable to transfer results");
-	for (i = 0; i < on; i++)
-	    r[i] = fr[i];
-    } else if (clEnqueueReadBuffer( commands, output, CL_TRUE, 0, sizeof(double) * on, REAL(res), 0, NULL, NULL ) != CL_SUCCESS)
-	Rf_error("Unable to transfer results");
+    /* we can release input memory objects now */
+    if (occ->mem_objects) {
+      arg_free(occ->mem_objects, (afin_t) clReleaseMemObject);
+      occ->mem_objects = 0;
+    }
+    if (float_args) {
+      arg_free(float_args, 0);
+      float_args = occ->float_args = 0;
+    }
 
-    clReleaseMemObject(output);
-    clReleaseCommandQueue(commands);
-    if (ftype == FT_SINGLE)
-	UNPROTECT(1);
+    res = ftres ? Rf_allocVector(RAWSXP, on * sizeof(float)) : Rf_allocVector(REALSXP, on);
+    if (ftype == FT_SINGLE) {
+	if (ftres) {
+	  if ((err = clEnqueueReadBuffer( commands, output, CL_TRUE, 0, sizeof(float) * on, RAW(res), 0, NULL, NULL )) != CL_SUCCESS)
+		Rf_error("Unable to transfer result vector (%d float elements, oclError %d)", on, err);
+	    PROTECT(res);
+	    Rf_setAttrib(res, R_ClassSymbol, mkString("clFloat"));
+	    UNPROTECT(1);
+	} else {
+	    /* float - need a temporary buffer */
+	    float *fr = (float*) malloc(sizeof(float) * on);
+	    double *r = REAL(res);
+	    int i;
+	    if (!fr)
+		Rf_error("unable to allocate memory for temporary single-precision output buffer");
+	    occ->float_out = fr;
+	    if ((err = clEnqueueReadBuffer( commands, output, CL_TRUE, 0, sizeof(float) * on, fr, 0, NULL, NULL )) != CL_SUCCESS)
+		Rf_error("Unable to transfer result vector (%d float elements, oclError %d)", on, err);
+	    for (i = 0; i < on; i++)
+		r[i] = fr[i];
+	}
+    } else if ((err = clEnqueueReadBuffer( commands, output, CL_TRUE, 0, sizeof(double) * on, REAL(res), 0, NULL, NULL )) != CL_SUCCESS)
+	Rf_error("Unable to transfer result vector (%d double elements, oclError %d)", on, err);
+
+    ocl_call_context_fin(octx);
+    UNPROTECT(1);
     return res;
 }
